@@ -1,7 +1,7 @@
 /** @file flint_lock.cc
  * @brief Flint-compatible database locking.
  */
-/* Copyright (C) 2005,2006,2007,2008,2009,2010,2011,2012,2013,2014,2015 Olly Betts
+/* Copyright (C) 2005,2006,2007,2008,2009,2010,2011,2012,2013,2014,2015,2016,2017 Olly Betts
  *
  * This program is free software; you can redistribute it and/or
  * modify it under the terms of the GNU General Public License as
@@ -30,7 +30,7 @@
 #include <unistd.h>
 #include <cstdlib>
 #include <sys/types.h>
-#include <sys/socket.h>
+#include "safesyssocket.h"
 #include <sys/wait.h>
 #include <signal.h>
 #include <cstring>
@@ -40,9 +40,8 @@
 #include "errno_to_string.h"
 #include "omassert.h"
 
-#ifdef __CYGWIN__
-# include <cygwin/version.h>
-# include <sys/cygwin.h>
+#ifdef FLINTLOCK_USE_FLOCK
+# include <sys/file.h>
 #endif
 
 #include "xapian/error.h"
@@ -59,8 +58,52 @@ using namespace std;
 # endif
 #endif
 
+bool
+FlintLock::test() const
+{
+#if defined __WIN32__
+    if (hFile != INVALID_HANDLE_VALUE) return true;
+    // Doesn't seem to be possible to check if the lock is held without briefly
+    // taking the lock.
+    throw Xapian::UnimplementedError("Can't test lock without trying to take it");
+#elif defined FLINTLOCK_USE_FLOCK
+    if (fd != -1) return true;
+    // Doesn't seem to be possible to check if the lock is held without briefly
+    // taking the lock.
+    throw Xapian::UnimplementedError("Can't test lock without trying to take it");
+#else
+    if (fd != -1) return true;
+    int lockfd = open(filename.c_str(), O_WRONLY | O_CREAT | O_TRUNC | O_CLOEXEC, 0666);
+    if (lockfd < 0) {
+	// Couldn't open lockfile.
+	reason why = ((errno == EMFILE || errno == ENFILE) ? FDLIMIT : UNKNOWN);
+	throw_databaselockerror(why, filename, "Testing lock");
+    }
+
+    struct flock fl;
+    fl.l_type = F_WRLCK;
+    fl.l_whence = SEEK_SET;
+    fl.l_start = 0;
+    fl.l_len = 1;
+    fl.l_pid = 0;
+    while (fcntl(lockfd, F_GETLK, &fl) == -1) {
+	if (errno != EINTR) {
+	    // Translate known errno values into a reason code.
+	    int e = errno;
+	    close(lockfd);
+	    // F_GETLK isn't implemented by GNU Hurd, and always fails with
+	    // ENOSYS: https://bugs.debian.org/190367
+	    reason why = (e == ENOLCK || e == ENOSYS ? UNSUPPORTED : UNKNOWN);
+	    throw_databaselockerror(why, filename, "Testing lock");
+	}
+    }
+    close(lockfd);
+    return fl.l_type != F_UNLCK;
+#endif
+}
+
 FlintLock::reason
-FlintLock::lock(bool exclusive, string & explanation) {
+FlintLock::lock(bool exclusive, bool wait, string & explanation) {
     // Currently we only support exclusive locks.
     (void)exclusive;
     Assert(exclusive);
@@ -81,23 +124,59 @@ FlintLock::lock(bool exclusive, string & explanation) {
 #else
     const char *fnm = filename.c_str();
 #endif
+retry:
+    // FIXME: Use LockFileEx() for locking, which would allow proper blocking
+    // and also byte-range locking for when we implement MVCC.  But is there a
+    // way to interwork with the CreateFile()-based locking while doing so?
     hFile = CreateFile(fnm, GENERIC_WRITE, FILE_SHARE_READ,
 		       NULL, OPEN_ALWAYS, FILE_ATTRIBUTE_NORMAL, NULL);
     if (hFile != INVALID_HANDLE_VALUE) return SUCCESS;
-    if (GetLastError() == ERROR_ALREADY_EXISTS) return INUSE;
-    explanation = string();
-    return UNKNOWN;
-#elif defined __EMX__
-    APIRET rc;
-    ULONG ulAction;
-    rc = DosOpen((PCSZ)filename.c_str(), &hFile, &ulAction, 0, FILE_NORMAL,
-		 OPEN_ACTION_OPEN_IF_EXISTS | OPEN_ACTION_CREATE_IF_NEW,
-		 OPEN_SHARE_DENYWRITE | OPEN_ACCESS_WRITEONLY,
-		 NULL);
-    if (rc == NO_ERROR) return SUCCESS;
-    if (rc == ERROR_ACCESS_DENIED) return INUSE;
+    if (GetLastError() == ERROR_ALREADY_EXISTS) {
+	if (wait) {
+	    Sleep(1000);
+	    goto retry;
+	}
+	return INUSE;
+    }
     explanation = string();
     return UNKNOWN;
+#elif defined FLINTLOCK_USE_FLOCK
+    // This is much simpler than using fcntl() due to saner semantics around
+    // releasing locks when closing other descriptors on the same file (at
+    // least on platforms where flock() isn't just a compatibility wrapper
+    // around fcntl()).  We can't simply switch to this without breaking
+    // locking compatibility with previous releases, though it might be useful
+    // for porting to platforms without fcntl() locking.  Also, flock()
+    // apparently doesn't work over NFS - perhaps that's OK, but we should at
+    // least check the failure mode.
+    Assert(fd == -1);
+    int lockfd = open(filename.c_str(), O_WRONLY | O_CREAT | O_TRUNC | O_CLOEXEC, 0666);
+    if (lockfd < 0) {
+	// Couldn't open lockfile.
+	explanation.assign("Couldn't open lockfile: ");
+	errno_to_string(errno, explanation);
+	return ((errno == EMFILE || errno == ENFILE) ? FDLIMIT : UNKNOWN);
+    }
+
+    int op = LOCK_EX;
+    if (!wait) op |= LOCK_NB;
+    while (flock(lockfd, op) == -1) {
+	if (errno != EINTR) {
+	    // Lock failed - translate known errno values into a reason code.
+	    close(lockfd);
+	    switch (errno) {
+		case EWOULDBLOCK:
+		    return INUSE;
+		case ENOLCK:
+		    return UNSUPPORTED; // FIXME: what do we get for NFS?
+		default:
+		    return UNKNOWN;
+	    }
+	}
+    }
+
+    fd = lockfd;
+    return SUCCESS;
 #else
     Assert(fd == -1);
 #if defined F_SETFD && defined FD_CLOEXEC
@@ -130,7 +209,7 @@ FlintLock::lock(bool exclusive, string & explanation) {
 	fl.l_start = 0;
 	fl.l_len = 1;
 	fl.l_pid = 0;
-	while (fcntl(lockfd, F_OFD_SETLK, &fl) == -1) {
+	while (fcntl(lockfd, wait ? F_OFD_SETLKW : F_OFD_SETLK, &fl) == -1) {
 	    if (errno != EINTR) {
 		if (errno == EINVAL) {
 		    // F_OFD_SETLK not supported by this kernel.
@@ -257,7 +336,7 @@ no_ofd_support:
 	    fl.l_whence = SEEK_SET;
 	    fl.l_start = 0;
 	    fl.l_len = 1;
-	    while (fcntl(lockfd, F_SETLK, &fl) == -1) {
+	    while (fcntl(lockfd, wait ? F_SETLKW : F_SETLK, &fl) == -1) {
 		if (errno != EINTR) {
 		    // Lock failed - translate known errno values into a reason
 		    // code.
@@ -286,7 +365,7 @@ no_ofd_support:
 	    if (why != SUCCESS) _exit(0);
 	}
 
-	// Make sure we don't block unmount() of partition holding the current
+	// Make sure we don't block unmount of partition holding the current
 	// directory.
 	if (chdir("/") < 0) {
 	    // We can't usefully do anything in response to an error, so just
@@ -301,7 +380,9 @@ no_ofd_support:
 	execl("/bin/cat", "/bin/cat", static_cast<void*>(NULL));
 	// Emulate cat ourselves (we try to avoid this to reduce VM overhead).
 	char ch;
-	while (read(0, &ch, 1) != 0) { /* Do nothing */ }
+	while (read(0, &ch, 1) != 0) {
+	    /* Do nothing */
+	}
 	_exit(0);
     }
 
@@ -360,10 +441,10 @@ FlintLock::release() {
     if (hFile == INVALID_HANDLE_VALUE) return;
     CloseHandle(hFile);
     hFile = INVALID_HANDLE_VALUE;
-#elif defined __EMX__
-    if (hFile == NULLHANDLE) return;
-    DosClose(hFile);
-    hFile = NULLHANDLE;
+#elif defined FLINTLOCK_USE_FLOCK
+    if (fd < 0) return;
+    close(fd);
+    fd = -1;
 #else
     if (fd < 0) return;
     close(fd);
@@ -393,7 +474,7 @@ FlintLock::release() {
 void
 FlintLock::throw_databaselockerror(FlintLock::reason why,
 				   const string & db_dir,
-				   const string & explanation)
+				   const string & explanation) const
 {
     string msg("Unable to get write lock on ");
     msg += db_dir;
